function [ Iter,FunctionData ] = CPUFractals( Nx,Ny,NumIter,Threshold,JuliaVal,SetType,xStart,xStop,yStart,yStop )
% @brief Computes either Julia or Mandelbrot set with Given Params.
% CPU Based processing with option to cancel.
% @param Nx : Amount of points in the X - Dimension
% @param Ny: Amount of points in the Y - Dimension
% @param Ax: Image Axis Handle
% @param NumIter: Amount of Iterations to Process
% @param Threshold: Iteration threshold 
% @param JuliaVal: Julia Set value (Not used in case of Mandlbrot)
% @param SetType: Either Julia Set or MandelBrot Set
% @param xStart & xStop: X - Axis Limits of the Set
% @param yStart & yStop: Y - Axis Limits of the Set

global S;

if(isfield(S,'WaitBar'))
    if(isvalid(S.WaitBar))
        delete(S.WaitBar);
    end
end


    % GENERATE MESHGRID
    XGrid = zeros(Ny,Nx);
    YGrid = zeros(Ny,Nx);
    for i = 1:Ny
        for j = 1:Nx
            jx = xStart + (j-1)*(xStop - xStart)/(Nx - 1);
            jy = yStart + (i-1)*(yStop - yStart)/(Ny - 1);
            XGrid(i,j) = jx;
            YGrid(i,j) = jy;
        end
    end


if(strcmpi(SetType,'JULIA'))
    

    % ARRAY INIT
    Cm = repmat(JuliaVal,[Ny,Nx]);
    NumIterations = zeros(Ny,Nx);
    A = complex(XGrid,YGrid);

    % MAIN JULIA COMPUTE
    
    S.WaitBar = waitbar(0,'1','Name','Computing Julia Set ...','CreateCancelBtn','setappdata(gcbf,''canceling'',1)');
    setappdata(S.WaitBar,'canceling',0);
    
    try
        for  i = 1:NumIter
            %disp(['Processing: ',num2str(i),' / ',num2str(NumIter),'...']);
            if getappdata(S.WaitBar,'canceling')
                break
            end

            waitbar(i/NumIter,S.WaitBar,sprintf("%3.1f %%",100*i/NumIter));

            AN = A.*A + Cm;

            Indices = abs(AN) < Threshold;
            A(Indices) = AN(Indices);
            NumIterations(Indices == true) = i;

        end
    catch Exception
        % REMOVE WAITBAR ON EXCEPTION
        delete(S.WaitBar);
        rethrow(Exception)
    end
    
    % REMOVE WAITBAR
    delete(S.WaitBar);

    % FINAL ASSIGNMENT
    Iter = NumIterations;
    FunctionData = abs(A);
    
elseif(strcmpi(SetType,'MANDEL'))
    % ARRAYS INIT
    NumIterations = zeros(Ny,Nx);
    c = complex(XGrid ,YGrid);

    
    S.WaitBar = waitbar(0,'1','Name','Computing Mandelbrot Set ...','CreateCancelBtn','setappdata(gcbf,''canceling'',1)');
    setappdata(S.WaitBar,'canceling',0);
    
    try
        % MAIN COMPUTING
        x = zeros(Ny,Nx);
        y = zeros(Ny,Nx);
        zn = zeros(Ny,Nx);
        for  i = 1:NumIter
           %disp(['Processing : ',num2str(i),'/',num2str(NumIter),' ...']);
            if getappdata(S.WaitBar,'canceling')
                break
            end
            
            waitbar(i/NumIter,S.WaitBar,sprintf("%3.1f %%",100*i/NumIter));

            % Compute
            xn = x.*x - y.*y + real(c);
            yn = 2*x.*y + imag(c);

            % Indices
            zn = xn.*xn + yn.*yn;
            Indices = abs(zn) < Threshold;
            NumIterations(Indices == true) = i;


            % Update for next iteration

            y(Indices) = yn(Indices);
            x(Indices) = xn(Indices);

        end
    
    catch Exception
        % REMOVE WAITBAR ON EXCEPTION
        delete(S.WaitBar);
        rethrow(Exception)
    end
    
    % REMOVE WAITBAR
    delete(S.WaitBar);

    % FINAL ASSIGNMENT
    Iter = NumIterations;
    FunctionData = abs(zn);
    
else
    error('Unknown Set!');
end


end

