// --------------------------------
// ------ CUDA Include Files ------
#include "cuda_runtime.h"
#include "device_launch_parameters.h"
#include <cufft.h>

// ---------------------------------
// ------- C++ Include Files -------
#include <stdio.h>
#include <string>
#include <iostream>


// ------------------------------------
// ------- Matlab Include Files -------
#include <matrix.h>
#include <mex.h>

// -------------------------------
// ------- Library Includes ------
#include "Julia.h"


/* *************************
* @brief CUDA ERROR HANDLER
***************************/
#define HANDLE_ERROR( err ) (HandleError( err, __FILE__, __LINE__ ))
static void HandleError(cudaError_t err, const char *file, int line) {
	if (err != cudaSuccess) {
		std::string str = std::string("CUDA ERROR: %s in %s at line %d\n") + cudaGetErrorString(err) + std::string(file) + std::to_string(line);
		mexErrMsgTxt(str.c_str());
	}
}


/**********************************************************
* @brief Main Structure holding the Device Memory Pointers
*********************************************************/
struct {
	int Hx = 0;
	int Hy = 0;
	int Nx = 0;
	int Ny = 0;
	int NewX = 0;
	int NewY = 0;
	int *Dev_Iterations = NULL;
	cufftComplex *Dev_Help0 = NULL;
	cufftComplex *Dev_Help1 = NULL;
	cufftComplex *Dev_Plane = NULL;
	cufftReal    *Dev_X = NULL;
	cufftReal    *Dev_Y = NULL;
	cufftReal	*Dev_Kernel = NULL;
	cufftHandle  FFTPlan = NULL;
	

	dim3 THREADS_PER_BLOCK = dim3(1, 1, 1);
	dim3 BLOCKS_PER_GRID = dim3(1, 1, 1);
}Vars;



/**********************************************************************
 * @brief Filters the Iteration Data with a custom Image Filter
 * @param Host_Kernel: Data Stored at CPU Representing the Image Kernel
 * @param Host_Data: Array for the Results
 **********************************************************************/
void ImageFilter(float *Host_Kernel,float*Host_Data, ComplexOutput CType){
	// COPY IMAGE KERNEL TO DEVICE
	cudaMemcpy(Vars.Dev_Help0, Host_Kernel, Vars.Hx * Vars.Hy * sizeof(cufftReal), cudaMemcpyHostToDevice);
	HANDLE_ERROR(cudaGetLastError());

	// ------------------------------------------------------
	// ------- CONVERT FROM COLUMN-MAJOR TO ROW-MAJOR -------
	Col2Row << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> >((cufftReal*)Vars.Dev_Help0, Vars.Dev_Kernel, Vars.Hx, Vars.Hy);
	HANDLE_ERROR(cudaGetLastError());

	// ---------------------------------------
	// ------- RESET GPU ARRAY TO ZERO -------
	cudaMemset(Vars.Dev_Help0, 0, Vars.NewX*Vars.NewY * sizeof(cufftComplex));
	HANDLE_ERROR(cudaGetLastError());

	// ----------------------------------------------------------
	// ------- DEPLOY KERNEL TO CENTER OF THE LARGE IMAGE -------
	DeployKernel << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> >(Vars.Dev_Kernel, Vars.Dev_Help0, Vars.Nx, Vars.Ny, Vars.Hx, Vars.Hy);
	HANDLE_ERROR(cudaGetLastError());

	// ------------------------------------------------------
	// ------- FORWARD FFT OF THE IMAGE FILTER KERNEL -------
	cufftResult Res = cufftExecC2C(Vars.FFTPlan, Vars.Dev_Help0, Vars.Dev_Help0, CUFFT_FORWARD);
	HANDLE_ERROR(cudaGetLastError());

	// ------------------------------------------
	// --> Spectrum of Filter Stored at Dev_Help0
	// ------------------------------------------

	// ---------------------------------------------
	// ------- CONVERT ITERATIONS TO COMPLEX -------
	IntToComplex << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> >(Vars.Dev_Iterations, Vars.Dev_Help1, Vars.NewX, Vars.NewY);
	HANDLE_ERROR(cudaGetLastError());

	// -----------------------------------------------
	// ------- FORWARD FFT OF THE IMAGE ITSELF -------
	Res = cufftExecC2C(Vars.FFTPlan, Vars.Dev_Help1, Vars.Dev_Help1, CUFFT_FORWARD);
	HANDLE_ERROR(cudaGetLastError());

	// -------------------------------------------------------
	// ------- COMPLEX MULTIPLY IN THE FREQUNCY DOMAIN -------
	ComplexMultiply << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> >(Vars.Dev_Help0, Vars.Dev_Help1, Vars.NewX, Vars.NewY);
	HANDLE_ERROR(cudaGetLastError());

	// ---------------------------------------
	// ------ INVERSE FFT OF THE RESULT ------
	Res = cufftExecC2C(Vars.FFTPlan, Vars.Dev_Help1, Vars.Dev_Help0, CUFFT_INVERSE);
	HANDLE_ERROR(cudaGetLastError());

	// OPTIONALLY NORMALIZE  AFTER IFFT - NOT IMPLEMENTED

	//IFFTSHIFT 2D
	int XHold = 0;
	int YHold = 0;
	//X Update
	if (Vars.NewX % 2 == 0) { XHold = static_cast<int>(std::ceil(Vars.NewX / 2));
	} else { XHold = static_cast<int>(std::ceil(Vars.NewX / 2)) + 1;}

	//Y Update
	if (Vars.NewY % 2 == 0) {YHold = static_cast<int>(std::ceil(Vars.NewY / 2));
	} else {YHold = static_cast<int>(std::ceil(Vars.NewY / 2)) + 1;}

	// ---------------------------------------------------------------
	// ------ 2D IFFTSHIFT. CAN BE SIMPLIFIED,USED OLD ROUTINE -------
	ShiftRows2D << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> >(Vars.Dev_Help0, Vars.Dev_Help1, Vars.NewX, Vars.NewY, XHold, YHold, 0);
	HANDLE_ERROR(cudaGetLastError());
	ShiftCols2D << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> >(Vars.Dev_Help1, Vars.Dev_Help0, Vars.NewX, Vars.NewY, XHold, YHold, 0);
	HANDLE_ERROR(cudaGetLastError());

	// -----------------------------------------------------
	// ------- CROP IMAGE BACK TO STANDARD DIMENSION -------
	CropImage << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> >(Vars.Dev_Help0, Vars.Dev_Help1, Vars.Nx, Vars.Ny, Vars.Hx, Vars.Hy);
	HANDLE_ERROR(cudaGetLastError());

	// -------------------------------
	// ------- COMPLEX TO REAL -------
	ComplexToCustom << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> >(Vars.Dev_Help1, (cufftReal*)Vars.Dev_Help0, CType, Vars.Nx, Vars.Ny);
	HANDLE_ERROR(cudaGetLastError());

	// ------------------------------------------------------
	// ------- CONVERT FROM ROW-MAJOR TO COLUMN-MAJOR -------
	Row2Col << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> >((cufftReal*)Vars.Dev_Help0, (cufftReal*)Vars.Dev_Help1, Vars.Nx, Vars.Ny);
	HANDLE_ERROR(cudaGetLastError());

	// ----------------------------------------------
	// ------- COPY RESULTS BACK INTO THE CPU -------
	cudaMemcpy(Host_Data, (cufftReal*)Vars.Dev_Help1, Vars.Nx * Vars.Ny * sizeof(cufftReal), cudaMemcpyDeviceToHost);
	HANDLE_ERROR(cudaGetLastError());
}


/******************************************************
 *@brief: Matlab Entry Function 
 * prhs = Pointer to The Right Hand Side	==> INPUT
 * plhs = Pointer to The Left Hand Side		==> OUTPUT
 ******************************************************/
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
	
	const mxArray *SetType		= NULL; // MandelBrot / Julia
	const mxArray *ComplexType	= NULL; // Complex / Real / Abs ...
	const mxArray *SetParams	= NULL; // Set Parameters
	const mxArray *ImageKernel	= NULL; // Custom Image Kernel
	if (nrhs > 0) { SetType = prhs[0]; }
	if (nrhs > 1) { ComplexType = prhs[1]; }
	if (nrhs > 2) { SetParams = prhs[2]; }
	if (nrhs > 3) { ImageKernel = prhs[3]; }

	// --------------------------------------
	// ------- CleanUp Used Resources -------
	if (nrhs == 1) {
		char *Buffer1 = new char[100];
		mxGetString(SetType, Buffer1, 100);
		std::string Type = std::string(Buffer1);

		if (_strcmpi(Type.c_str(), "EXIT") == 0) {
			if (Vars.FFTPlan != NULL)			{ cufftDestroy(Vars.FFTPlan); }
			if (Vars.Dev_Plane != NULL)			{ cudaFree(Vars.Dev_Plane); Vars.Dev_Plane = NULL; }
			if (Vars.Dev_X != NULL)				{ cudaFree(Vars.Dev_X); Vars.Dev_X = NULL; }
			if (Vars.Dev_Y != NULL)				{ cudaFree(Vars.Dev_Y); Vars.Dev_Y = NULL; }
			if (Vars.Dev_Iterations != NULL)	{ cudaFree(Vars.Dev_Iterations); Vars.Dev_Iterations = NULL; }
			if (Vars.Dev_Help0 != NULL)			{ cudaFree(Vars.Dev_Help0); Vars.Dev_Help0 = NULL; }
			if (Vars.Dev_Help1 != NULL)			{ cudaFree(Vars.Dev_Help1); Vars.Dev_Help1 = NULL; }
			if (Vars.Dev_Kernel != NULL)		{ cudaFree(Vars.Dev_Kernel); Vars.Dev_Kernel = NULL; }
			delete[] Buffer1;
			Vars.Nx = 1;
			Vars.Ny = 1;
			return;
		}

	}

	char *Buffer1 = new char[100];
	char *Buffer2 = new char[100];
	mxGetString(SetType, Buffer1, 100);
	mxGetString(ComplexType, Buffer2, 100);
	std::string Type = std::string(Buffer1);
	std::string CType_s = std::string(Buffer2);
	delete[] Buffer1;
	delete[] Buffer2;

	// COMPUTING PARAMETERS 
	int Nx = 0;
	int Ny = 0;
	int Hx = 0;
	int Hy = 0;

	// ---------------------------------------------------
	// ------ GET IMAGE KERNEL (Filter) DIMENSIONS -------
	float *KernelData = NULL;
	if (ImageKernel != NULL) {
		mxClassID ID = mxGetClassID(ImageKernel);
		if(ID != mxSINGLE_CLASS) {
			mexErrMsgTxt("ECC ERROR: Image Kernel need to be single-precision!\n");
		}

		const mwSize *Dimensions = mxGetDimensions(ImageKernel);
		
		Hx = static_cast<int>(Dimensions[1]);
		Hy = static_cast<int>(Dimensions[0]);
		KernelData = static_cast<float*>(mxGetData(ImageKernel));
	}

	float xStart = 0;
	float yStart = 0;
	float xStop = 0;
	float yStop = 0;
	int ProcessImage = 0;
	
	float Threshold = 0;
	int NumIter = 0;
	int ChunkIter = 0;
	float JuliaReal = 0.0;
	float JuliaImag = 0.0;
	ComputingType SType = JULIA;
	ComplexOutput CType = REAL;
	if (mxGetClassID(SetParams) != mxDOUBLE_CLASS) {
		mexErrMsgTxt("ERROR: Expecting Input Parameters to be double!\n");
	}

	double *Parameters = (double*)mxGetData(SetParams);
	


	if (_strcmpi(CType_s.c_str(), "real") == 0) {
		CType = REAL;
	} else if (_strcmpi(CType_s.c_str(), "imag") == 0) {
		CType = IMAG;
	} else if (_strcmpi(CType_s.c_str(), "abs") == 0) {
		CType = ABS;
	} else if (_strcmpi(CType_s.c_str(), "angle") == 0) {
		CType = ANGLE;
	} else {
		mexErrMsgTxt("Unknown Complex Transformation \n");
	}

	// ---------------------------------------------------------------------------------------------------------------
	if (_strcmpi(Type.c_str(), "Julia") == 0) { // ------- ------ ------- Julia Set Processing ------- ------- -------
	// ---------------------------------------------------------------------------------------------------------------
		if (mxGetNumberOfElements(SetParams) < 12) {
			mexPrintf("Requires: \n");
			mexPrintf("Nx : Amount of Elements in the X - Dimension\n");
			mexPrintf("Ny: Amount of Elements in the Y - Dimension\n");
			mexPrintf("xStart : X - Axis Start Point\n");
			mexPrintf("xStop  : X - Axis End   Point\n");
			mexPrintf("yStart : Y - Axis Start Point\n");
			mexPrintf("yStop  : Y - Axis End Point \n");
			mexPrintf("Threshold: Iteration Threshold\n");
			mexPrintf("NumIter  : Amount of Iterations to Process\n ");
			mexPrintf("ChunkIter: Amount of Iterations per kernel call\n");
			mexPrintf("Zoom | Axis Scaling \n");
			mexPrintf("Julia real(C)\n");
			mexPrintf("Julia imag(C)\n");
			mexPrintf("ProcessImage using Kernel\n");
			mexErrMsgTxt("Please Check the above!\n");
		}

		Nx				= static_cast<int>(Parameters[0]);
		Ny				= static_cast<int>(Parameters[1]);
		xStart			= static_cast<float>(Parameters[2]);
		xStop			= static_cast<float>(Parameters[3]);
		yStart			= static_cast<float>(Parameters[4]);
		yStop			= static_cast<float>(Parameters[5]);
		Threshold		= static_cast<float>(Parameters[6]);
		NumIter			= static_cast<int>(Parameters[7]);
		ChunkIter		= static_cast<int>(Parameters[8]);
		JuliaReal		= static_cast<float>(Parameters[9]);
		JuliaImag		= static_cast<float>(Parameters[10]);
		ProcessImage	= static_cast<int>(Parameters[11]);

		SType = JULIA;
	// -------------------------------------------------------------------------------------------------------------------------
	} else if (_strcmpi(Type.c_str(), "MANDEL") == 0) { // ------- ------- ------- MandelBrot Processing ------- ------- -------
	// -------------------------------------------------------------------------------------------------------------------------
		if (mxGetNumberOfElements(SetParams) < 10) {
			mexPrintf("Requires: \n");
			mexPrintf("Nx : Amount of Elements in the X - Dimension\n");
			mexPrintf("Ny: Amount of Elements in the Y - Dimension\n");
			mexPrintf("xStart : X - Axis Start Point\n");
			mexPrintf("xStop  : X - Axis End   Point\n");
			mexPrintf("yStart : Y - Axis Start Point\n");
			mexPrintf("yStop  : Y - Axis End Point \n");
			mexPrintf("Threshold: Iteration Threshold\n");
			mexPrintf("NumIter  : Amount of Iterations to Process\n ");
			mexPrintf("ChunkIter: Amount of Iterations per kernel call\n");
			mexPrintf("Zoom | Axis Scaling \n");
			mexPrintf("ProcessImage using Kernel\n");
			mexErrMsgTxt("Please Check the above!\n");
		}

		Nx				= static_cast<int>(Parameters[0]);
		Ny				= static_cast<int>(Parameters[1]);
		xStart			= static_cast<float>(Parameters[2]);
		xStop			= static_cast<float>(Parameters[3]);
		yStart			= static_cast<float>(Parameters[4]);
		yStop			= static_cast<float>(Parameters[5]);
		Threshold		= static_cast<float>(Parameters[6]);
		NumIter			= static_cast<int>(Parameters[7]);
		ChunkIter		= static_cast<int>(Parameters[8]);
		ProcessImage	= static_cast<int>(Parameters[9]);
		SType = MANDEL;

	} else {
		mexErrMsgTxt("Invalid Method!\n");
		return;
	}

	// ----------------------------------------------------------------------------
	// ------- Reallocate GPU Memory only if any relevant Parameter Changes -------
	// NOTE: This is to speedup Execution
	// NX ... Image Size X
	// NY ... Image Size Y
	// HX ... Filter Kernel X
	// HY ... Filter Kernel Y
	cudaGetLastError();
	if ((Vars.Nx != Nx) || (Vars.Ny != Ny) || (Vars.Hx != Hx) || (Vars.Hy != Hy)) {

		Vars.Nx = Nx; Vars.Hx = Hx;
		Vars.Ny = Ny; Vars.Hy = Hy;
		Vars.NewX = Nx + 2*Hx;
		Vars.NewY = Ny + 2*Hy;

		// --------------------------------------
		// ------ PLAN CUDA FFT Execution -------
		cufftPlan2d(&Vars.FFTPlan, Vars.NewY, Vars.NewX, CUFFT_C2C);

		// ---------------------------------------
		// ------- Clean All Device Memory -------
		if (Vars.Dev_Plane != NULL) {
			cudaFree(Vars.Dev_Plane);
			Vars.Dev_Plane = NULL;
			HANDLE_ERROR(cudaGetLastError());
		}
		if (Vars.Dev_X != NULL) {
			cudaFree(Vars.Dev_X);
			Vars.Dev_X = NULL;
			HANDLE_ERROR(cudaGetLastError());
		}
		if (Vars.Dev_Y != NULL) {
			cudaFree(Vars.Dev_Y);
			Vars.Dev_Y = NULL;
			HANDLE_ERROR(cudaGetLastError());
		}
		if (Vars.Dev_Help0 != NULL) {
			cudaFree(Vars.Dev_Help0);
			Vars.Dev_Help0 = NULL;
			HANDLE_ERROR(cudaGetLastError());
		}

		if (Vars.Dev_Help1 != NULL) {
			cudaFree(Vars.Dev_Help1);
			Vars.Dev_Help1 = NULL;
			HANDLE_ERROR(cudaGetLastError());
		}

		if (Vars.Dev_Iterations != NULL) {
			cudaFree(Vars.Dev_Iterations);
			Vars.Dev_Iterations = NULL;
			HANDLE_ERROR(cudaGetLastError());
		}

		if (Vars.Dev_Kernel != NULL) {
			cudaFree(Vars.Dev_Kernel);
			Vars.Dev_Kernel = NULL;
			HANDLE_ERROR(cudaGetLastError());
		}

		// ---------------------------------
		// ------- Setup Thread Grid -------
		Vars.THREADS_PER_BLOCK = dim3(16, 16, 1);
		Vars.BLOCKS_PER_GRID   = dim3(Vars.NewX / 16 + 1, Vars.NewY / 16 + 1, 1);

		// -----------------------------------
		// ------- Allocate GPU Memory -------
		cudaError_t Err = cudaMalloc((void**)&Vars.Dev_Plane, Vars.NewX * Vars.NewY * sizeof(cufftComplex));
		if (Err != cudaSuccess) {
			if (Vars.Dev_Plane != NULL) {
				cudaFree(Vars.Dev_Plane);
				Vars.Dev_Plane = NULL;
			}
			HANDLE_ERROR(cudaGetLastError());
		}

		if ((Vars.Hx > 0) && (Vars.Hy > 0)) {
			Err = cudaMalloc((void**)&Vars.Dev_Kernel, Vars.Hx * Vars.Hy * sizeof(cufftReal));
			if (Err != cudaSuccess) {
				if (Vars.Dev_Kernel != NULL) {
					cudaFree(Vars.Dev_Kernel);
					Vars.Dev_Kernel = NULL;
				}
				HANDLE_ERROR(cudaGetLastError());
			}
		}

		Err = cudaMalloc((void**)&Vars.Dev_Help0, Vars.NewX * Vars.NewY * sizeof(cufftComplex));
		if (Err != cudaSuccess) {
			if (Vars.Dev_Help0 != NULL) {
				cudaFree(Vars.Dev_Help0);
				Vars.Dev_Help0 = NULL;
			}
			HANDLE_ERROR(cudaGetLastError());
		}

		Err = cudaMalloc((void**)&Vars.Dev_Help1, Vars.NewX * Vars.NewY * sizeof(cufftComplex));
		if (Err != cudaSuccess) {
			if (Vars.Dev_Help1 != NULL) {
				cudaFree(Vars.Dev_Help1);
				Vars.Dev_Help1 = NULL;
			}
			HANDLE_ERROR(cudaGetLastError());
		}

		Err = cudaMalloc((void**)&Vars.Dev_X, Vars.NewX * Vars.NewY * sizeof(cufftReal));
		if (Err != cudaSuccess) {
			if (Vars.Dev_X != NULL) {
				cudaFree(Vars.Dev_X);
				Vars.Dev_X = NULL;
			}
			HANDLE_ERROR(cudaGetLastError());
		}
		Err = cudaMalloc((void**)&Vars.Dev_Y, Vars.NewX * Vars.NewY * sizeof(cufftReal));
		if (Err != cudaSuccess) {
			if (Vars.Dev_Y != NULL) {
				cudaFree(Vars.Dev_Y);
				Vars.Dev_Y = NULL;
			}
			HANDLE_ERROR(cudaGetLastError());
		}
		Err = cudaMalloc((void**)&Vars.Dev_Iterations, Vars.NewX * Vars.NewY * sizeof(int));
		if (Err != cudaSuccess) {
			if (Vars.Dev_Iterations != NULL) {
				cudaFree(Vars.Dev_Iterations);
				Vars.Dev_Iterations = NULL;
			}
			HANDLE_ERROR(cudaGetLastError());
		}
	}


	// ---------------------------------------
	// ------- Check Memory Allocation -------
	if ((Vars.Dev_Plane == NULL) || (Vars.Dev_X == NULL) || (Vars.Dev_Y == NULL) || (Vars.Dev_Iterations == NULL)) {
		mexErrMsgTxt("CUDA Error Memory Allocation\n");
	}

	// PREPARE MAIN ARRAYS 
	JuliaMandelInitialize << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> > (Vars.Dev_Plane, xStart, xStop, yStart, yStop, Vars.NewX, Vars.NewY);
	HANDLE_ERROR(cudaGetLastError());

	int NumLoops = static_cast<int>(ceil(static_cast<float>(NumIter) / static_cast<float>(ChunkIter)));

	// -----------------------------------------
	// ------- Reset Used Memory Buffers -------
	cudaMemset(Vars.Dev_X, 0, Vars.NewX * Vars.NewY * sizeof(cufftReal));
	HANDLE_ERROR(cudaGetLastError());
	cudaMemset(Vars.Dev_Y, 0, Vars.NewX * Vars.NewY * sizeof(cufftReal));
	HANDLE_ERROR(cudaGetLastError());
	cudaMemset(Vars.Dev_Help0, 0, Vars.NewX * Vars.NewY * sizeof(cufftReal));
	HANDLE_ERROR(cudaGetLastError());
	cudaMemset(Vars.Dev_Help1, 0, Vars.NewX * Vars.NewY * sizeof(cufftReal));
	HANDLE_ERROR(cudaGetLastError());
	cudaMemset(Vars.Dev_Iterations, 0, Vars.NewX * Vars.NewY * sizeof(int));
	HANDLE_ERROR(cudaGetLastError());
	cudaMemset(Vars.Dev_Kernel, 0, Vars.Hx * Vars.Hy * sizeof(cufftReal));
	HANDLE_ERROR(cudaGetLastError());

	// ----------------------------------------
	// ------- Grab Julia Complex Value -------
	cufftComplex JuliaConst;
	JuliaConst.x = JuliaReal;
	JuliaConst.y = JuliaImag;

	// ---------------------------------------------------
	// ------- ------- Main Computing Loop ------- -------
	// ---------------------------------------------------
	int IterationsMissing = NumIter;
	for (int i = 0; i < NumLoops; i++) {

		int IterationsToProcess = ChunkIter;
		if (IterationsMissing < ChunkIter) { IterationsToProcess = IterationsMissing; } // Last Piece
		if (IterationsToProcess == 0) { break; }                                      // In case Last piece is 0

		if (SType == JULIA) {
			JuliaKernel << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> >(Vars.Dev_Plane, JuliaConst, Vars.Dev_Iterations, Threshold, IterationsToProcess, Vars.NewX, Vars.NewY);
			HANDLE_ERROR(cudaGetLastError());
		} else {
			MandelBrotKernel << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> >(Vars.Dev_Plane, Vars.Dev_X, Vars.Dev_Y, Vars.Dev_Iterations, Threshold, IterationsToProcess, Vars.NewX, Vars.NewY);
			HANDLE_ERROR(cudaGetLastError());
		}

		// Update for Next Round
		IterationsMissing -= ChunkIter;
	}


	// ----------------------------------------------
	// ------- Prepare Matlab Output Buffers --------
	const mxArray* Iterations   = plhs[0] = mxCreateNumericMatrix(Ny, Nx, mxSINGLE_CLASS, mxREAL);
	const mxArray* FunctionData = plhs[1] = mxCreateNumericMatrix(Ny, Nx, mxSINGLE_CLASS, mxREAL);
	float *M_Data = (float*)mxGetData(FunctionData); // Convergence
	float *M_Iter = (float*)mxGetData(Iterations);   // Iterations

	switch (SType) {
		// ----------------------------------------------------------------------------------------
		case JULIA: { // ------ ------- ------- Finish Processing JULIA Set ------- ------- -------
		// ----------------------------------------------------------------------------------------
		// 
			// CROP IMAGE (Julia Function Values) To Dev_Help
			CropImage << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> > (Vars.Dev_Plane, Vars.Dev_Help0, Vars.Nx, Vars.Ny, Vars.Hx, Vars.Hy);
			HANDLE_ERROR(cudaGetLastError());

			// TAKE REAL OTUPUT of the Function Values to Dev_X
			ComplexToCustom << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> > (Vars.Dev_Help0, (cufftReal*)Vars.Dev_Help1, CType, Vars.Nx, Vars.Ny);
			HANDLE_ERROR(cudaGetLastError());

			// CONVERT FROM ROW MAJOR TO COLUMN MAJOR
			Row2Col << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> > ((cufftReal*)Vars.Dev_Help1, (cufftReal*)Vars.Dev_Help0, Vars.Nx, Vars.Ny);
			HANDLE_ERROR(cudaGetLastError());

			// COPY CROPPED FUNCTION VALUES INTO THE CPU MEMORY
			cudaMemcpy(M_Data, Vars.Dev_Help0, Vars.Nx* Vars.Ny * sizeof(cufftReal), cudaMemcpyDeviceToHost);
			HANDLE_ERROR(cudaGetLastError());

			// ------------------------------------------------------------------------------------
			if (ProcessImage) { // ------- POST PROCESS ITERATIONS BY A CUSTOM IMAGE FILTER -------
			// ------------------------------------------------------------------------------------

				// It is required to filter the image,but filter has not been included in the call
				if ((KernelData == NULL) || (ImageKernel == NULL)) {
					mexErrMsgTxt("Requested Additional Image Processing without Kernel!\n");
				}

				// FILTER IMAGE WITH A CUSTOM KERNEL VIA FFT
				ImageFilter(KernelData, M_Iter, CType);

			// -----------------------------------------------------------------------
			}else { // ------ DO NOT FILTER THE ITERATIONS BY A CUSTOM KERNEL -------
			// -----------------------------------------------------------------------

				// CONVERT INTEGER TO FLOAT
				IntToFloat << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> > (Vars.Dev_Iterations, Vars.Dev_X, Vars.NewX, Vars.NewY);
				HANDLE_ERROR(cudaGetLastError());

				// CROP THE IMAGE BACK TO NORMAL SIZE
				CropImage << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> > (Vars.Dev_X, Vars.Dev_Y, Vars.Nx, Vars.Ny, Vars.Hx, Vars.Hy);
				HANDLE_ERROR(cudaGetLastError());

				// CONVERT FROM ROW-MAJOR TO COLUMN-MAJOR
				Row2Col << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> > (Vars.Dev_Y, Vars.Dev_X, Vars.Nx, Vars.Ny);
				HANDLE_ERROR(cudaGetLastError());


				// COPY AMOUNT OF ITERATIONS INTO THE CPU
				cudaMemcpy(M_Iter, Vars.Dev_X, Vars.Nx * Vars.Ny * sizeof(cufftReal), cudaMemcpyDeviceToHost);
				HANDLE_ERROR(cudaGetLastError());
			}
			break;
		}
		// ---------------------------------------------------------------------------------------------- 
		case MANDEL: { // ------ ------- ------- Finish Processing MANDELBROT Set ------- ------- -------
		// ---------------------------------------------------------------------------------------------- 
		// 
			// TAKE REAL OTUPUT
			ComplexToCustom << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> > (Vars.Dev_X, Vars.Dev_Y, (cufftReal*)Vars.Dev_Help0, CType, Vars.NewX, Vars.NewY);
			HANDLE_ERROR(cudaGetLastError());

			// CROP IMAGE BACK TO STANDARD RESOLUTION
			CropImage << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> > ((cufftReal*)Vars.Dev_Help0, (cufftReal*)Vars.Dev_Help1, Vars.Nx, Vars.Ny, Vars.Hx, Vars.Hy);
			HANDLE_ERROR(cudaGetLastError());


			// CONVERT FROM ROW-MAJOR TO COLUMN - MAJOR
			Row2Col << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> > ((cufftReal*)Vars.Dev_Help1, (cufftReal*)Vars.Dev_Help0, Vars.Nx, Vars.Ny);
			HANDLE_ERROR(cudaGetLastError());

			// COPY FUNCTION VALUES INTO THE CPU MEMORY
			cudaMemcpy(M_Data, Vars.Dev_Help0, Vars.Nx * Vars.Ny * sizeof(cufftReal), cudaMemcpyDeviceToHost);
			HANDLE_ERROR(cudaGetLastError());

			// ------------------------------------------------------------------------------------
			if (ProcessImage) { // ------- POST PROCESS ITERATIONS BY A CUSTOM IMAGE FILTER -------
			// ------------------------------------------------------------------------------------

				// It is required to filter the image,but filter has not been included in the call
				if ((KernelData == NULL) || (ImageKernel == NULL)) {
					mexErrMsgTxt("Requested Additional Image Processing without Kernel!\n");
				}

				// FILTER IMAGE WITH A CUSTOM KERNEL VIA FFT
				ImageFilter(KernelData, M_Iter, CType);

			// ----------------------------------------------------------------------
			}else { // ------ DO NOT FILTER THE ITERATIONS BY A CUSTOM KERNEL -------
			// ----------------------------------------------------------------------

				// CONVERT INTEGER TO FLOAT AND TRANSPOSE ARRAY - USE UNUSED COMPLEX BUFFER
				IntToFloat << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> > (Vars.Dev_Iterations, (cufftReal*)Vars.Dev_Help0, Vars.NewX, Vars.NewY);
				HANDLE_ERROR(cudaGetLastError());

				// CROP TO NORMAL RESOLUTION
				CropImage << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> > ((cufftReal*)Vars.Dev_Help0, (cufftReal*)Vars.Dev_Help1, Vars.Nx, Vars.Ny, Vars.Hx, Vars.Hy);
				HANDLE_ERROR(cudaGetLastError());

				// FROM ROW MAJOR TO COLUMNS MAJOR
				Row2Col << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> > ((cufftReal*)Vars.Dev_Help1, (cufftReal*)Vars.Dev_Help0, Vars.Nx, Vars.Ny);
				HANDLE_ERROR(cudaGetLastError());

				// COPY AMOUNT OF ITERATIONS INTO THE CPU
				cudaMemcpy(M_Iter, Vars.Dev_Help0, Vars.Nx * Vars.Ny * sizeof(cufftReal), cudaMemcpyDeviceToHost);
				HANDLE_ERROR(cudaGetLastError());

			}
			break;
		}
	}



}