// --------------------------------
// ------ CUDA Include Files ------
#include "cuda_runtime.h"
#include "device_launch_parameters.h"
#include <cufft.h>

// ------------------------------
// ------ Library Includes ------
#include "Julia.h"

/**************************************************
* @brief Implements Complex Multiply on the Device
**************************************************/
__device__ cufftComplex operator* (const cufftComplex& lhs, const cufftComplex& rhs) {
	cufftComplex Temp;
	Temp.x = (lhs.x * rhs.x - lhs.y * rhs.y);
	Temp.y = (lhs.x * rhs.y + lhs.y * rhs.x);
	return Temp;
}

/**************************************************
* @brief Implements Complex Addition on the Device
**************************************************/
__device__ cufftComplex operator + (const cufftComplex& lhs, const cufftComplex& rhs) {
	cufftComplex Temp;
	Temp.x = lhs.x + rhs.x;
	Temp.y = lhs.y + rhs.y;
	return Temp;
}

/*******************************************************************
*@brief Implements Absolute Value of a complex number on the device
******************************************************************/
__device__ cufftReal abs(cufftComplex Val) {
	return sqrtf(Val.x * Val.x + Val.y * Val.y);
}

/****************************************************************
* @brief Main Julia Kernel
* @param Plane: Julia Function Values
* @param C: Custom Julia Set Complex Constant
* @param Iterations: Amount of iterations processed for each point
* @param Threshold: Threshold for computing
* @param NumIter: Amount of Iterations per kernel call
* @param Nx X Dimensioon of the Grid
* @param Ny Y Dimension of the Grid
*****************************************************************/
__global__ void JuliaKernel(cufftComplex* Plane, cufftComplex C, int* Iterations, float Threshold, int NumIter, int Nx, int Ny) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	if (tix < Nx && tiy < Ny) {
		int tid = tix + tiy * Nx;

		// Load from global memory
		cufftComplex LA = Plane[tid];
		int IterProcessed = Iterations[tid];

		// Chunk Iteration Processing Loop
		for (unsigned int i = 0; i < NumIter; i++) {
			cufftComplex Res = LA * LA + C;
			cufftReal Absolute = abs(Res);

			if (Absolute < Threshold) {
				LA = Res;
				IterProcessed++;
			}
		}

		// Upload to Global Memory
		Plane[tid] = LA;
		Iterations[tid] = IterProcessed;
	}

}

/***********************************************************
* @brief Converts Integer Array into Double Precision Array
***********************************************************/
__global__ void IntToFloat(int* Data_In, cufftReal* Data_Out, int Nx, int Ny) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	if (tix < Nx && tiy < Ny) {
		int tid = tix + tiy * Nx;
		Data_Out[tid] = (float)(Data_In[tid]);
	}
}

/******************************************************************************************
*@brief Transforms input Integer Data into complex Data filling imaginary part with zeros.
******************************************************************************************/
__global__ void IntToComplex(int* Data_In, cufftComplex* Data_Out, int Nx, int Ny) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	if ((tix < Nx) && (tiy < Ny)) {
		int tid = tix + tiy * Nx;
		Data_Out[tid].x = (float)Data_In[tid];
		Data_Out[tid].y = 0.0f;
	}
}


/**************************************************
 * @brief Takes only The Real out of Complex Number
 *************************************************/
__global__ void ComplexToReal(cufftComplex* Data_In, cufftReal* Data_Out, int Nx, int Ny) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	if (tix < Nx && tiy < Ny) {
		int tid = tix + tiy * Nx;
		Data_Out[tid] = Data_In[tid].x;
	}
}

/********************************
* @brief Matrix Complex multiply
*******************************/
__global__ void ComplexMultiply(cufftComplex* In, cufftComplex* Out, int Nx, int Ny) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	if ((tix < Nx) && (tiy < Ny)) {
		int tid = tix + tiy * Nx;
		Out[tid] = In[tid] * Out[tid];
	}
}

/******************************************************
* @brief Deploys Kernel to the FFT Center
* Hx,Hy : X & Y - Dimensions of the Image Filter Kernel
*******************************************************/
__global__ void DeployKernel(cufftReal* Kernel, cufftComplex* Data, int Nx, int Ny, int Hx, int Hy) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	int NewX = Nx + 2 * Hx;
	int NewY = Ny + 2 * Hy;


	if (tix < Hx && tiy < Hy) {
		int tid = tix + tiy * Hx;
		int xStart = floorf((Nx + Hx) / 2.0f);
		int yStart = floorf((Ny + Hy) / 2.0f);
		int xid = tix + xStart;
		int yid = tiy + yStart;
		int wid = xid + yid * NewX;
		Data[wid].x = Kernel[tid];
		Data[wid].y = 0.0f;
	}
}

/*******************************************************************************************
*  @brief Crops the Image so that the Kernel artifacts are not visible at the image borders
*******************************************************************************************/
__global__ void CropImage(cufftComplex* BigImage, cufftComplex* SmallImage, int Nx, int Ny, int Hx, int Hy) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;

	int NewX = Nx + 2 * Hx;
	int NewY = Ny + 2 * Hy;

	if ((tix >= Hx) && (tix < (Hx + Nx)) && (tiy >= Hy) && (tiy < (Hy + Ny))) {
		int tid = tix + tiy * NewX;
		int nxid = tix - Hx;
		int nyid = tiy - Hy;
		int wid = nxid + nyid * Nx;
		SmallImage[wid] = BigImage[tid];
	}

}

/*******************************************************************************************
*  @brief Crops the Image so that the Kernel artifacts are not visible at the image borders
*******************************************************************************************/
__global__ void CropImage(cufftReal* BigImage, cufftReal* SmallImage, int Nx, int Ny, int Hx, int Hy) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;

	int NewX = Nx + 2 * Hx;
	int NewY = Ny + 2 * Hy;

	if ((tix >= Hx) && (tix < (Hx + Nx)) && (tiy >= Hy) && (tiy < (Hy + Ny))) {
		int tid = tix + tiy * NewX;
		int nxid = tix - Hx;
		int nyid = tiy - Hy;
		int wid = nxid + nyid * Nx;
		SmallImage[wid] = BigImage[tid];
	}
}

/**************************************************************
* @brief Computes the Mandebrot set
* @param C - Mandelbrot Constant
* @param X - Coordinates
* @param Y - Coordinates
* @param Iterations - Iterations processed for a given point
* @param Threshold - Iteration Threshold
* @param NumIter - Amount of Chunk Iterations
* @param Nx - Amount of Points in the X - Dimension
* @param Ny - Amount of points in the Y - Dimension
***************************************************************/
__global__ void MandelBrotKernel(cufftComplex* C, cufftReal* X, cufftReal* Y, int* Iterations, float Threshold, int NumIter, int Nx, int Ny) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	if (tix < Nx && tiy < Ny) {
		int tid = tix + tiy * Nx;

		// LOAD FROM GLOBAL MEMORY
		cufftComplex Cl = C[tid];
		cufftReal Xl = X[tid];
		cufftReal Yl = Y[tid];

		// COMPUTE CHUNK
		int IterProcessed = Iterations[tid];
		for (unsigned int i = 0; i < NumIter; i++) {
			cufftReal Xn = Xl * Xl - Yl * Yl + Cl.x;
			cufftReal Yn = 2 * Xl * Yl + Cl.y;
			cufftReal Res = Xn * Xn + Yn * Yn;
			if (Res < Threshold) {
				IterProcessed++;
				Yl = Yn;
				Xl = Xn;
			}
		}

		// UPLOAD RESULTS TO GLOBAL MEMORY
		X[tid] = Xl; Y[tid] = Yl;
		Iterations[tid] = IterProcessed;

	}
}

/**********************************************************
* @brief Initializes the Julia And MandelBrot Computations
* Basically a 2D linspace for X and Y Corrdinates
*********************************************************/
__global__ void JuliaMandelInitialize(cufftComplex* Plane, float xStart, float xStop, float yStart, float yStop, int Nx, int Ny) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	if (tix < Nx && tiy < Ny) {
		int tid = tix + tiy * Nx;
		cufftComplex Temp;
		Temp.x = xStart + tix * (xStop - xStart) / (Nx - 1);
		Temp.y = yStart + tiy * (yStop - yStart) / (Ny - 1);
		Plane[tid] = Temp;
	}
}


/************************************************************
* @brief Changes the Column Major Format to Row Major Format
************************************************************/
__global__ void Col2Row(cufftReal* in, cufftReal* out, int Nx, int Ny) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	if ((tix < Nx) && (tiy < Ny)) {
		int tin = tix * Ny + tiy;
		int tou = tiy * Nx + tix;
		out[tou] = in[tin];
	}
}


/****************************************************************
*  @ brief Converts from ROW-major format to COLUMN-major format
***************************************************************/
__global__ void Row2Col(cufftReal* Data_In, cufftReal* Data_Out, int Nx, int Ny) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	if (tix < Nx && tiy < Ny) {
		int tin = tiy * Nx + tix;
		int tou = tix * Ny + tiy;
		Data_Out[tou] = Data_In[tin];
	}
}

/************************************************************
* @brief Takes Compex Input and Transforms it to Real Output
* Also Does Operation Column to Row Order
************************************************************/
__global__ void ComplexToCustom(cufftComplex* DIn, cufftReal* DOut, ComplexOutput Out, int Nx, int Ny) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	if (tix < Nx && tiy < Ny) {
		int tid = tiy * Nx + tix;

		cufftComplex C = DIn[tid];
		if (Out == REAL) {
			DOut[tid] = C.x;
		}
		else if (Out == IMAG) {
			DOut[tid] = C.y;
		}
		else if (Out == ABS) {
			DOut[tid] = sqrtf(C.x * C.x + C.y * C.y);
		}
		else if (Out == ANGLE) {
			DOut[tid] = atanf(C.y / C.x);
		}

	}
}

// Shift data along the X - Dimension
__global__ void ShiftRows2D(cufftComplex* in, cufftComplex* out, int Nx, int Ny, int XHold, int YHold, int dir) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	if ((tix < Nx) && (tiy < Ny)) {

		//On which side is our thread
		if (tix < XHold) {
			int readIndex = tiy * Nx + tix;
			int writeIndex = tiy * Nx + (Nx - XHold + tix);
			out[writeIndex] = in[readIndex];

		}
		else {

			//Minor update
			if (dir == 1) {//FFTSHIFT
						   //Minor update
				if (Nx % 2 == 0) {
					tix = tix - (Nx - XHold);
				}
				else {
					tix = tix - (Nx - XHold + 1);
				}
			}
			else {//IFFTSHIFT
				if (Nx % 2 == 0) {
					tix = tix - (Nx - XHold);
				}
				else {
					tix = tix - (Nx - XHold - 1);
				}

			}
			int readIndex = tiy * Nx + tix + XHold;
			int writeIndex = tiy * Nx + tix;
			out[writeIndex] = in[readIndex];
		}
	}
}

//Shift data along the Y - dimension
__global__ void ShiftCols2D(cufftComplex* in, cufftComplex* out, int Nx, int Ny, int XHold, int YHold, int dir) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	if ((tix < Nx) && (tiy < Ny)) {

		//On which side is our thread
		if (tiy < YHold) {
			int readIndex = tiy * Nx + tix;
			int writeIndex = tix + (Ny - YHold + tiy) * Nx;
			out[writeIndex] = in[readIndex];
		}
		else {
			//Minor Update
			if (dir == 1) { //FFTSHIFT
				if (Ny % 2 == 0) {
					tiy = tiy - (Ny - YHold);
				}
				else {
					tiy = tiy - (Ny - YHold + 1);
				}
			}
			else {//IFFTSHIFT

				if (Ny % 2 == 0) {
					tiy = tiy - (Ny - YHold);
				}
				else {
					tiy = tiy - (Ny - YHold - 1);
				}


			}
			int readIndex = (tiy + YHold) * Nx + tix;
			int writeIndex = tiy * Nx + tix;
			out[writeIndex] = in[readIndex];
		}
	}
}


/************************************************************
* @brief Takes Compex Input and Transforms it to Real Output
* Also Does Operation Column to Row Order
************************************************************/
__global__ void ComplexToCustom(cufftReal* DIn_R, cufftReal* DIn_I, cufftReal* DOut, ComplexOutput Out, int Nx, int Ny) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	if (tix < Nx && tiy < Ny) {
		int tid = tiy * Nx + tix;

		cufftComplex C;
		C.x = DIn_R[tid];
		C.y = DIn_I[tid];
		if (Out == REAL) {
			DOut[tid] = C.x;
		}
		else if (Out == IMAG) {
			DOut[tid] = C.y;
		}
		else if (Out == ABS) {
			DOut[tid] = sqrtf(C.x * C.x + C.y * C.y);
		}
		else if (Out == ANGLE) {
			DOut[tid] = atanf(C.y / C.x);
		}

	}
}
